/** Copyright (C) 2017 by Bundesanstalt für Gewässerkunde
 * Software engineering by
 *  Björnsen Beratende Ingenieure GmbH
 *  Dr. Schumacher Ingenieurbüro für Wasser und Umwelt
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.artifacts.bundu.bezugswst;

import java.util.ArrayList;
import java.util.List;

import org.dive4elements.artifacts.CallContext;
import org.dive4elements.river.artifacts.access.FixRealizingAccess;
import org.dive4elements.river.artifacts.bundu.BUNDUArtifact;
import org.dive4elements.river.artifacts.bundu.BunduResultType;
import org.dive4elements.river.artifacts.common.AbstractResultType;
import org.dive4elements.river.artifacts.common.GeneralResultType;
import org.dive4elements.river.artifacts.common.ResultRow;
import org.dive4elements.river.artifacts.model.Calculation;
import org.dive4elements.river.artifacts.model.Calculation.Problem;
import org.dive4elements.river.artifacts.model.CalculationResult;
import org.dive4elements.river.artifacts.model.DateRange;
import org.dive4elements.river.artifacts.model.WQKms;
import org.dive4elements.river.artifacts.model.fixings.FixRealizingCalculation;
import org.dive4elements.river.artifacts.model.fixings.FixRealizingResult;
import org.dive4elements.river.artifacts.model.river.RiverInfoProvider;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.artifacts.services.DynamicMainValuesTimeRangeDeterminationService;
import org.dive4elements.river.artifacts.services.DynamicMainValuesTimeRangeDeterminationService.GaugeInfoResult;
import org.dive4elements.river.artifacts.sinfo.tkhstate.BedHeightsFinder;
import org.dive4elements.river.artifacts.sinfo.tkhstate.BedQualityD50TimeRangeConfig;
import org.dive4elements.river.artifacts.sinfo.tkhstate.WinfoArtifactWrapper;
import org.dive4elements.river.artifacts.sinfo.util.CalculationUtils;
import org.dive4elements.river.artifacts.sinfo.util.RiverInfo;
import org.dive4elements.river.artifacts.sinfo.util.WstInfo;
import org.dive4elements.river.artifacts.states.WaterlevelData;
import org.dive4elements.river.exports.WaterlevelDescriptionBuilder;
import org.dive4elements.river.model.BedHeightValueType;
import org.dive4elements.river.model.River;
import org.dive4elements.river.utils.DoubleUtil;

class BezugswstCalculation {

    // private static Logger log = Logger.getLogger(BezugswstCalculation.class);

    /**
     * Additional depth (m) to compute the excavation volume
     */
    private final static double EXCAVATION_DEPTH = 0.2; // REMARK Sollte von außen einstellbar sein

    /**
     * Excavation costs (euro) per cubic meter
     */
    private final static double EXPENSE_PER_CBM = 12.0; // REMARK Sollte von außen einstellbar sein

    private final static double KM_TO_M = 1000.0;

    private final CallContext context;

    private final List<ResultRow> rows;

    private Double missKmFrom;

    private Double missKmTo;

    public BezugswstCalculation(final CallContext context) {
        this.context = context;
        this.rows = new ArrayList<>();
    }

    /**
     * Calculates the result rows of a bundu bzws workflow
     */
    public CalculationResult calculate(final BUNDUArtifact bunduartifact) {

        // Get input data
        final String user = CalculationUtils.findArtifactUser(this.context, bunduartifact);
        final BunduAccess access = new BunduAccess(bunduartifact);
        final River river = access.getRiver();
        final RiverInfo riverInfo = new RiverInfo(river);
        final String calcModeLabel = Resources.getMsg(this.context.getMeta(), "bundu_bezugswst");
        final boolean preprocessing = access.getPreprocessing();
        final int startYear = access.getStartYear();
        final int endYear = access.getBezugsJahr();
        final Integer ud = access.getUd();
        this.missKmFrom = access.getMissingVolFrom();
        this.missKmTo = access.getMissingVolTo();

        final GaugeInfoResult gi = DynamicMainValuesTimeRangeDeterminationService
                .getCommonTimeRangeForGauges(river.determineGauges(access.getLowerKm(), access.getUpperKm()), startYear, endYear, this.context.getMeta());
        final int globalAdjustedEndYear = gi.getGlobalEndYear();
        final int globalAdjustedStartYear = gi.getGlobalStartYear();

        final BezugswstCalculationResults results = new BezugswstCalculationResults(calcModeLabel, user, riverInfo, access.getRange(),
                access.isCalculateMissingVolume());

        final Calculation problems = new Calculation();

        // Calculate the wspl for the selected river range as in fixa awspl
        bunduartifact.addStringData("wq_isq", "true");
        final WinfoArtifactWrapper winfo = new WinfoArtifactWrapper(bunduartifact);
        final RiverInfoProvider riverInfoProvider = RiverInfoProvider.forRange(this.context, river, access.getRange());
        final FixRealizingResult fixResult = calculateWspl(bunduartifact, problems);
        if (fixResult == null)
            return new CalculationResult(results, problems);

        final WQKms wqkms = fixResult.getWQKms()[0];
        // We have no wst year as the wst is created by a calculation; we do not need it though
        final int wspYear = -1;
        // Remark: showAllGauges true for Fixierungsanalyse, false for WInfo, so true here as well
        final boolean showAllGauges = true;
        final WaterlevelData waterlevel = new WaterlevelData(wqkms, wspYear, showAllGauges, true);
        final RiverInfoProvider riverInfoProvider2 = riverInfoProvider.forWaterlevel(waterlevel);
        final WstInfo wstInfo = new WstInfo(wqkms.getName(), 0, riverInfoProvider2.getReferenceGauge(), true);

        // Fetch the bed levels of the selected sounding
        final Integer bedHeightId = access.getBedHeightID();
        final BedHeightsFinder bedHeightsFinder = (bedHeightId != null) ? BedHeightsFinder.forId(problems, bedHeightId, access.getRange(), false)
                : BedHeightsFinder.NullFinder();

        // Fetch the river channel data
        final ChannelFinder channelFinder = ChannelFinder.loadValues(problems, river, access.getBezugsJahr());
        if (channelFinder == null)
            return new CalculationResult(results, problems);

        // Compute the result rows
        for (int i = 0; i <= wqkms.size() - 1; i++) {
            this.rows.add(createRow(wqkms.getKm(i), wqkms.getW(i), wqkms.getQ(i), bedHeightsFinder, channelFinder, riverInfoProvider2, wstInfo));
        }

        // Compute the missing volumes
        if (access.isCalculateMissingVolume()) {
            if ((bedHeightsFinder == null) || bedHeightsFinder.isNull())
                return new CalculationResult(results, problems);
            computeMissingVolumes(problems);
            final BedQualityCalculator bqCalculator = computeDensities(problems, bunduartifact, access, river);
            if (bqCalculator != null)
                computeMissingMasses(problems, bqCalculator);
        }

        // Add the result to the results collection
        final WaterlevelDescriptionBuilder descBuilder = new WaterlevelDescriptionBuilder(winfo, this.context);
        final String qtext = descBuilder.getMetadataQ();
        final BezugswstMainCalculationResult result = new BezugswstMainCalculationResult("bundu-bzws", this.rows, bedHeightsFinder.getInfo(), wstInfo,
                access.getFunction(), preprocessing, globalAdjustedStartYear, globalAdjustedEndYear, ud, qtext, wqkms, this.missKmFrom, this.missKmTo);
        results.addResult(result, problems);

        // Create the missing volume results
        if (access.getMissingVolFrom() != null) {
            final String title1 = Resources.getMsg(this.context.getMeta(), "bundu.export.csv.title.bezugswst.result1");
            final BezugswstMissVolCalculationResult1 r1 = new BezugswstMissVolCalculationResult1(title1, this.rows);
            results.addResult(r1, null);

            final String title2 = Resources.getMsg(this.context.getMeta(), "bundu.export.csv.title.bezugswst.result2");
            final List<ResultRow> rows2 = copyMissRows();
            final BezugswstMissVolCalculationResult2 r2 = new BezugswstMissVolCalculationResult2(title2, rows2);
            results.addResult(r2, null);

            final String title3 = Resources.getMsg(this.context.getMeta(), "bundu.export.csv.title.bezugswst.result3");
            final List<ResultRow> totalRows = new ArrayList<>();
            totalRows.add(createTotalsRow(problems));
            final BezugswstMissVolCalculationResult3 r3 = new BezugswstMissVolCalculationResult3(title3, totalRows);
            results.addResult(r3, null);
        }

        return new CalculationResult(results, problems);
    }

    /**
     * Calculates a w-q-longitudinal section for a river range and Q specified in an artifact
     */
    private FixRealizingResult calculateWspl(final BUNDUArtifact bundu, final Calculation problems) {

        final FixRealizingAccess access = new FixRealizingAccess(bundu);
        final FixRealizingCalculation calc = new FixRealizingCalculation(access);

        final CalculationResult res = calc.calculate(this.context.getMeta());

        final FixRealizingResult fixRes = (FixRealizingResult) res.getData();

        final List<Problem> problems2 = res.getReport().getProblems();
        if (problems2 != null) {
            for (final Problem problem : problems2) {
                problems.addProblem(problem);
            }
        }
        return fixRes;
    }

    /**
     * Create a result row for a station
     */
    private ResultRow createRow(final double station, final double w, final double q, final BedHeightsFinder bedHeightsFinder,
            final ChannelFinder channelFinder, final RiverInfoProvider riverInfoProv, final WstInfo wstInfo) {

        // Set W and Q
        final ResultRow row = ResultRow.create();
        row.putValue(GeneralResultType.station, station);
        row.putValue(BunduResultType.bezugswst, w);
        row.putValue(GeneralResultType.dischargeQwithUnit, q);
        row.putValue(GeneralResultType.waterlevelLabel, wstInfo.getLabel());
        row.putValue(GeneralResultType.gaugeLabel, riverInfoProv.findGauge(station));
        row.putValue(GeneralResultType.location, riverInfoProv.getLocation(station));
        if (bedHeightsFinder.getInfo() != null)
            row.putValue(BunduResultType.sounding, bedHeightsFinder.getInfo().getDescription());
        else
            row.putValue(BunduResultType.sounding, "");

        // Set bed and channel bottom height
        final double msh = bedHeightsFinder.getMeanBedHeight(station);
        row.putValue(BunduResultType.heightMeanBed, msh);
        if (!Double.isNaN(w) && !Double.isNaN(msh))
            row.putValue(BunduResultType.flowdepthMeanBed, w - msh);
        else
            row.putValue(BunduResultType.flowdepthMeanBed, Double.NaN);

        final double channelDepth = channelFinder.getDepth(station);
        row.putValue(BunduResultType.channelDepth, channelDepth);
        double channelHeight;
        if (!Double.isNaN(w) && !Double.isNaN(channelDepth))
            channelHeight = w - channelDepth;
        else
            channelHeight = Double.NaN;
        row.putValue(BunduResultType.channelLowerEdge, channelHeight);
        final double channelWidth = channelFinder.getWidth(station);
        row.putValue(BunduResultType.channelWidth, channelWidth);
        if (!Double.isNaN(channelHeight)) {
            if (Double.isNaN(msh))
                row.putValue(BunduResultType.missDepthMeanBed, Double.NaN);
            else if (msh > channelHeight + 0.001)
                row.putValue(BunduResultType.missDepthMeanBed, msh - channelHeight);
            else
                row.putValue(BunduResultType.missDepthMeanBed, 0.0);
        }

        // Set field heights and missing heights
        final List<Double> fieldHeights = new ArrayList<>();
        final List<Double> fieldDepths = new ArrayList<>();
        final List<Double> fieldMissDepths = new ArrayList<>();
        final List<Double> fieldMissWidths = new ArrayList<>();
        final List<Double> fieldNulls = new ArrayList<>();
        int missFieldCnt = 0;
        for (int i = BedHeightValueType.FIELD_FIRST_INDEX; i <= BedHeightValueType.FIELD_LAST_INDEX; i++) {
            final double h = bedHeightsFinder.getFieldHeight(station, i);
            fieldHeights.add(Double.valueOf(h));
            fieldDepths.add(Double.valueOf(w - h));
            if (Double.isNaN(h)) {
                fieldMissDepths.add(Double.NaN);
                fieldMissWidths.add(Double.NaN);
            }
            else if (h > channelHeight + 0.001) {
                missFieldCnt++;
                fieldMissDepths.add(Double.valueOf(h - channelHeight));
                fieldMissWidths.add(Double.valueOf(channelWidth / BedHeightValueType.FIELD_LAST_INDEX));
            } else {
                fieldMissDepths.add(Double.valueOf(0.0));
                fieldMissWidths.add(Double.valueOf(0.0));
            }
            fieldNulls.add(Double.NaN);
        }
        if (!Double.isNaN(msh) && isKmInMissingVolumeRange(station)) {
            row.putValue(BunduResultType.missDepthFields, fieldMissDepths);
            row.putValue(BunduResultType.missWidthFields, fieldMissWidths);
            row.putValue(BunduResultType.hasMissingDepth, (missFieldCnt >= 1));
        } else {
            row.putValue(BunduResultType.missDepthFields, fieldNulls);
            row.putValue(BunduResultType.missWidthFields, fieldNulls);
            row.putValue(BunduResultType.hasMissingDepth, null);
        }
        row.putValue(BunduResultType.missVolumeFields, fieldNulls);
        row.putValue(BunduResultType.missMassFields, fieldNulls);
        row.putValue(BunduResultType.bedHeightFields, fieldHeights);
        row.putValue(BunduResultType.depthFields, fieldDepths);

        // Preset the missing volume fields with NaN
        row.putValue(BunduResultType.excavationCosts, Double.NaN);
        row.putValue(BunduResultType.excavationVolume, Double.NaN);
        row.putValue(BunduResultType.missVolumeMeanBed, Double.NaN);
        row.putValue(BunduResultType.missMassMeanBed, Double.NaN);
        row.putValue(BunduResultType.missVolumeTotal, Double.NaN);
        row.putValue(BunduResultType.missMassTotal, Double.NaN);
        row.putValue(BunduResultType.density, Double.NaN);
        row.putValue(BunduResultType.missStationRangeFrom, Double.NaN);
        row.putValue(BunduResultType.missStationRangeTo, Double.NaN);

        return row;
    }

    /**
     * Computes the missing volumes in a km range
     */
    private void computeMissingVolumes(final Calculation problems) {
        // Search start km
        int first = -1;
        for (int j = 0; j <= this.rows.size() - 1; j++) {
            if (isKmInMissingVolumeRange(this.rows.get(j).getDoubleValue(GeneralResultType.station))) {
                first = j;
                break;
            }
        }
        if (first < 0)
            return;
        // Calculate all kms in missing volume calc range
        double km;
        int last = this.rows.size() - 1;
        for (int i = first; i <= this.rows.size() - 1; i++) {
            km = this.rows.get(i).getDoubleValue(GeneralResultType.station);
            if (!isKmInMissingVolumeRange(km))
                break;
            if (km > this.missKmTo.doubleValue() - 0.0001)
                last = i;
            if (this.rows.get(i).getValue(BunduResultType.hasMissingDepth) == null)
                continue;
            final double chDepth = this.rows.get(i).getDoubleValue(BunduResultType.channelDepth) + EXCAVATION_DEPTH;
            final List<Double> areas = new ArrayList<>();
            final List<Double> volumes = new ArrayList<>();
            double vTotal = 0.0;
            double vExcav = 0.0;
            for (int j = BedHeightValueType.FIELD_FIRST_INDEX; j <= BedHeightValueType.FIELD_LAST_INDEX; j++) {
                if (getFieldValue(i, BunduResultType.missDepthFields, j) > 0.0001) {
                    computeMissingVolume(volumes, areas, i, first, last, j, ActualMissingHeightComputer.Instance);
                    vTotal += volumes.get(j - 1);
                } else {
                    volumes.add(Double.valueOf(0.0));
                    areas.add(Double.valueOf(0.0));
                }
                if (chDepth - getFieldValue(i, BunduResultType.depthFields, j) > 0.0001) {
                    vExcav += computeMissingVolume(null, null, i, first, last, j, ExcavationMissingHeightComputer.Instance);
                }
            }
            final double[] meanBedVolumeArea = computeMeanBedMissingAreaAndVolume(i, first, last);
            this.rows.get(i).putValue(BunduResultType.missVolumeMeanBed, meanBedVolumeArea[0]);
            this.rows.get(i).putValue(BunduResultType.missAreaMeanBed, meanBedVolumeArea[1]);
            this.rows.get(i).putValue(BunduResultType.missVolumeFields, volumes);
            this.rows.get(i).putValue(BunduResultType.missAreaFields, areas);
            this.rows.get(i).putValue(BunduResultType.missVolumeTotal, vTotal);
            this.rows.get(i).putValue(BunduResultType.excavationVolume, vExcav);
            this.rows.get(i).putValue(BunduResultType.excavationCosts, vExcav * EXPENSE_PER_CBM);
        }
    }

    /**
     * Computes the missing volume of a field of a km row
     */
    private double computeMissingVolume(final List<Double> volumes, final List<Double> areas, final int current, final int first, final int last,
            final int field, final MissingHeightComputer heightcomputer) {

        final double dhCurr = heightcomputer.missingHeight(this.rows.get(current), current, first, last, field);
        final double dhPrev = heightcomputer.missingHeight(this.rows.get(current - 1), current - 1, first, last, field);
        final double dhNext = heightcomputer.missingHeight(this.rows.get(current + 1), current + 1, first, last, field);
        final double kmCurr = kmOfRow(current);
        final double kmPrev = kmOfRow(current - 1);
        final double kmNext = kmOfRow(current + 1);
        final double width = getFieldValue(current, BunduResultType.missWidthFields, field);
        final double area1 = (0.25 * dhPrev + 0.75 * dhCurr) * width;
        final double dist1 = Double.isNaN(kmPrev) ? 0.0 : Math.abs(kmCurr - kmPrev) * KM_TO_M / 2;
        final double area2 = (0.75 * dhCurr + 0.25 * dhNext) * width;
        final double dist2 = Double.isNaN(kmNext) ? 0.0 : Math.abs(kmNext - kmCurr) * KM_TO_M / 2;
        final double volume = dist1 * area1 + dist2 * area2;
        if (volumes != null)
            volumes.add(volume);
        if (areas != null) {
            if (!Double.isNaN(volume))
                areas.add(Double.valueOf(area1 + area2));
            else
                areas.add(Double.NaN);
        }
        return volume;
    }

    /**
     * Interface for the function that computes/gets the missing height of a field
     */
    private interface MissingHeightComputer {
        /**
         * Gets the missing height of a field and a row if in range, otherwise 0.0
         */
        double missingHeight(final ResultRow row, final int rowIndex, final int first, final int last, final int fieldIndex);
        /**
         * Gets the mean missing height of a row if in range, otherwise 0.0
         */
        double missingMeanHeight(final ResultRow row, final int rowIndex, final int first, final int last);
    }

    /**
     * Computation of the actual missing height of a field
     */
    private static class ActualMissingHeightComputer implements MissingHeightComputer {
        public static MissingHeightComputer Instance = new ActualMissingHeightComputer();

        /**
         * Gets the missing height of a field and a row if in range, otherwise 0.0
         */
        @SuppressWarnings("unchecked")
        @Override
        public double missingHeight(final ResultRow row, final int rowIndex, final int first, final int last, final int fieldIndex) {
            if ((first <= rowIndex) && (rowIndex <= last)) {
                final double dh = ((List<Double>) row.getValue(BunduResultType.missDepthFields)).get(fieldIndex - 1).doubleValue();
                return (!Double.isNaN(dh) ? dh : 0.0);
            }
            else
                return 0.0;
        }

        /**
         * Gets the missing mean height of a row if in range, otherwise 0.0
         */
        @Override
        public double missingMeanHeight(final ResultRow row, final int rowIndex, final int first, final int last) {
            if ((first <= rowIndex) && (rowIndex <= last)) {
                final double dh = row.getDoubleValue(BunduResultType.missDepthMeanBed);
                return (!Double.isNaN(dh) ? dh : 0.0);
            }
            else
                return 0.0;
        }
    }

    /**
     * Computation of the excavation height of a field
     */
    private static class ExcavationMissingHeightComputer implements MissingHeightComputer {
        public static MissingHeightComputer Instance = new ExcavationMissingHeightComputer();

        /**
         * Gets the excavation height of a field and a row if in range, otherwise 0.0
         */
        @SuppressWarnings("unchecked")
        @Override
        public double missingHeight(final ResultRow row, final int rowIndex, final int first, final int last, final int fieldIndex) {
            if ((first <= rowIndex) && (rowIndex <= last)) {
                final double channeldepth = row.getDoubleValue(BunduResultType.channelDepth) + EXCAVATION_DEPTH;
                final double fielddepth = ((List<Double>) row.getValue(BunduResultType.depthFields)).get(fieldIndex - 1).doubleValue();
                return (!Double.isNaN(channeldepth - fielddepth) ? Math.max(channeldepth - fielddepth, 0.0) : 0.0);
            }
            else
                return 0.0;
        }

        /**
         * Gets the excavation mean height of a row if in range, otherwise 0.0
         */
        @Override
        public double missingMeanHeight(final ResultRow row, final int rowIndex, final int first, final int last) {
            if ((first <= rowIndex) && (rowIndex <= last)) {
                final double channeldepth = row.getDoubleValue(BunduResultType.channelDepth) + EXCAVATION_DEPTH;
                final double flowdepth = row.getDoubleValue(BunduResultType.flowdepthMeanBed);
                return (!Double.isNaN(channeldepth - flowdepth) ? Math.max(channeldepth - flowdepth, 0.0) : 0.0);
            }
            else
                return 0.0;
        }
    }

    /**
     * Computes the missing area and volume of the mean bed level of a km row
     */
    private double[] computeMeanBedMissingAreaAndVolume(final int current, final int first, final int last) {

        final double dhCurr = ActualMissingHeightComputer.Instance.missingMeanHeight(this.rows.get(current), current, first, last);
        if (dhCurr < 0.0001)
            return new double[] { 0.0, 0.0 };
        final double dhPrev = ActualMissingHeightComputer.Instance.missingMeanHeight(this.rows.get(current - 1), current - 1, first, last);
        final double dhNext = ActualMissingHeightComputer.Instance.missingMeanHeight(this.rows.get(current + 1), current + 1, first, last);
        final double kmCurr = kmOfRow(current);
        final double kmPrev = kmOfRow(current - 1);
        final double kmNext = kmOfRow(current + 1);
        final double width = this.rows.get(current).getDoubleValue(BunduResultType.channelWidth);
        final double area1 = (0.25 * dhPrev + 0.75 * dhCurr) * width;
        final double dist1 = Double.isNaN(kmPrev) ? 0.0 : Math.abs(kmCurr - kmPrev) * KM_TO_M / 2;
        final double area2 = (0.75 * dhCurr + 0.25 * dhNext) * width;
        final double dist2 = Double.isNaN(kmNext) ? 0.0 : Math.abs(kmNext - kmCurr) * KM_TO_M / 2;
        final double volume = dist1 * area1 + dist2 * area2;
        final double area = Double.isNaN(volume) ? Double.NaN : area1 + area2;
        return new double[] { volume, area };
    }

    /**
     * Gets the km of a row index if within range, otherwise NaN
     */
    private double kmOfRow(final int rowIndex) {
        if ((0 <= rowIndex) && (rowIndex <= this.rows.size() - 1)) // && (this.rows.get(rowIndex).getValue(BunduResultType.hasMissingDepth) != null))
            return this.rows.get(rowIndex).getDoubleValue(GeneralResultType.station);
        return Double.NaN;
    }

    /**
     * Create a density calculator and compute the densities of the missing volume km range and time period according to the
     * reference year
     */
    private BedQualityCalculator computeDensities(final Calculation problems, final BUNDUArtifact bunduartifact, final BunduAccess access, final River river) {
        final BedQualityCalculator bqCalculator = new BedQualityCalculator(this.context, bunduartifact);
        // REMARK 10km tolerance at start and end to enable interpolation there
        final double[] kms = DoubleUtil.explode(access.getMissingVolFrom().doubleValue() - 10.0, access.getMissingVolTo().doubleValue() + 10.0,
                access.getStep().doubleValue() / 1000);
        final DateRange dateRange = BedQualityD50TimeRangeConfig.getDefaults(river, access.getBezugsJahr().intValue(), problems);
        if (dateRange == null)
            return null;
        bqCalculator.execute(problems, river, kms, dateRange.getFrom(), dateRange.getTo());
        return bqCalculator;
    }

    /**
     * Computes the missing masses
     */
    private void computeMissingMasses(final Calculation problems, final BedQualityCalculator densityFinder) {
        for (final ResultRow row : this.rows) {
            @SuppressWarnings("unchecked")
            final List<Double> volumes = (List<Double>) row.getValue(BunduResultType.missVolumeFields);
            if ((volumes == null) || Double.isNaN(volumes.get(0)))
                continue;
            final double density = getDensity(row.getDoubleValue(GeneralResultType.station), densityFinder);
            final List<Double> masses = new ArrayList<>();
            double kmTotal = 0.0;
            int mcnt = 0;
            for (int j = BedHeightValueType.FIELD_FIRST_INDEX; j <= BedHeightValueType.FIELD_LAST_INDEX; j++) {
                final double m = volumes.get(j - 1) * density;
                masses.add(m);
                if (!Double.isNaN(m)) {
                    kmTotal += m;
                    mcnt += 1;
                }
            }
            if (mcnt == 0)
                kmTotal = Double.NaN;
            row.putValue(BunduResultType.density, density);
            row.putValue(BunduResultType.missMassFields, masses);
            row.putValue(BunduResultType.missMassTotal, kmTotal);
            row.putValue(BunduResultType.missMassMeanBed, row.getDoubleValue(BunduResultType.missVolumeMeanBed) * density);
        }
    }

    /**
     * Gets a value of one of the field list types of a row
     *
     * @param rowIndex
     * @param type
     * @param fieldIndex
     *            1-based field index
     */
    private double getFieldValue(final int rowIndex, final AbstractResultType type, final int fieldIndex) {
        @SuppressWarnings("unchecked")
        final List<Double> values = (List<Double>) this.rows.get(rowIndex).getValue(type);
        return values.get(fieldIndex - 1);
    }

    /**
     * Computes the volume and mass total of all rows with missing volumes
     */
    private ResultRow createTotalsRow(final Calculation problems) {
        // Search start km
        double vTotal = 0.0;
        double mTotal = 0.0;
        double eTotal = 0.0;
        double cTotal = 0.0;
        int vcnt = 0;
        int mcnt = 0;
        int ecnt = 0;
        for (final ResultRow row : this.rows) {
            final double volume = row.getDoubleValue(BunduResultType.missVolumeTotal);
            final double mass = row.getDoubleValue(BunduResultType.missMassTotal);
            if (!Double.isNaN(volume)) {
                vTotal += volume;
                vcnt++;
                if (!Double.isNaN(mass)) {
                    mTotal += mass;
                    mcnt++;
                }
            }
            final double excavation = row.getDoubleValue(BunduResultType.excavationVolume);
            final double costs = row.getDoubleValue(BunduResultType.excavationCosts);
            if (!Double.isNaN(excavation)) {
                eTotal += excavation;
                cTotal += costs;
                ecnt++;
            }
        }
        if (vcnt == 0)
            vTotal = Double.NaN;
        if (mcnt == 0)
            mTotal = Double.NaN;
        if (ecnt == 0) {
            eTotal = Double.NaN;
            cTotal = Double.NaN;
        }
        final ResultRow sumRow = ResultRow.create();
        sumRow.putValue(BunduResultType.missStationRangeFrom, Double.valueOf(this.missKmFrom));
        sumRow.putValue(BunduResultType.missStationRangeTo, Double.valueOf(this.missKmTo));
        sumRow.putValue(BunduResultType.missVolumeTotal, vTotal);
        sumRow.putValue(BunduResultType.missMassTotal, mTotal);
        sumRow.putValue(BunduResultType.excavationVolumeTotal, eTotal);
        sumRow.putValue(BunduResultType.excavationCostsTotal, cTotal);
        return sumRow;
    }

    /**
     * Copies the rows of the missing volume calculation range into a new list
     */
    private List<ResultRow> copyMissRows() {
        final List<ResultRow> missRows = new ArrayList<>();
        for (final ResultRow row : this.rows) {
            final double km = row.getDoubleValue(GeneralResultType.station);
            if (isKmInMissingVolumeRange(km))
                missRows.add(row);
        }
        return missRows;
    }

    /**
     * Gets the density of a km from the densities calculation
     */
    private double getDensity(final double km, final BedQualityCalculator densityFinder) {
        return densityFinder.getDensity(km);
    }

    /**
     * Checks whether a km lies in the missing volume calculation range
     */
    private boolean isKmInMissingVolumeRange(final double km) {
        if ((this.missKmFrom == null) || (this.missKmTo == null))
            return false;
        return (this.missKmFrom.doubleValue() - 0.0001 < km) && (km < this.missKmTo.doubleValue() + 0.0001);
    }
}